# Hashing for Cuckoo Filters


### What hashes do we need?

 - h_1: the "location hash" for our filter
	 - Input: string; stored as array of characters (one of the words in our dictionary)
	 - Output: number from 0 to number of slots in our array
	 - We'll have a large number output and then take modulo
  
 - f: the "fingerprint hash" for our filter
	 - Input: string; stored as array of characters (one of the words in our dictionary)
	 - Output: number from 1 to 255
	 - We'll have a large number output and then take modulo (making sure it starts at 1)
 
 - h: the way to hash fingerprints to a larger universe (for XORing with h_1)
	 - Input: fingerprint from 1 to 255
	 - Output: number from 1 to number of slots in our array
	 - We'll have a large number output and then take modulo (making sure it starts at 1)



### What can we use?

We've seen multiply-shift hashing in Mini-Midterm 3 (let's take a quick look).  Can that work? 
 - Issue #1: works on integers, not strings
 - Issue #2: Guarantees in terms of items hashing to the same or similar values aren't great 
	 - (we'll discuss this more later)

MurmurHash: a popular, open-source hash
 - No theoretical guarantees
 - But works really well in practice; very fast
 - Takes strings as input; gives 128 bits as output
 - One of many popular options for this task





### Using MurmurHash

A few things we need to bear in mind when setting up MurmurHash.

First, MurmurHash requires a seed.  
We want to keep the seed consistent throughout the run
 - that way, MurmurHash outputs the same value for a given string

My suggestion:
 - Choose a large random number and use it as a seed
 - For debugging, you probably want this number to be a constant
 - But in general, it's safe to choose a random number (using rand()) as seed

Second, MurmurHash outputs 128 bits each time it's called
 - This works well for us!  Can use the first 64 bits for h_1, and the second 64 bits for f
 - Allocate a two-element array for MurmurHash to store its output in
 - This array is small, so it's OK to put it on the stack

With this, we have a method for h_1 and f.  Let's take a look at how this works.





### What about h?

Need a method to hash fingerprints to large numbers
 - We can assume fingerprints have values 1...255

Idea: when the universe of possible elements is so small, hashing is easy
 - Can just store all hash values

hashFingerprint[] is a 255 element array full of large numbers
(let's take a look at filter.c)

Can hash fingerprint with value i using hashFingerprint[i-1]
 - Array lookup is WAY faster (and easier) than MurmurHash

Then take mod to get output into the correct range
 - Remember that h should always output a value >=1!



# Quick lab demo

## How starter version looks (no filter)

## Correct final version (hopefully with speedup)

## How to test


# Good luck with your lab!
